# Repository path:    $HeadURL: https://90.145.156.222/DMS-Houtribdijk/code/nc-to-thredds/source/golf_MEMpuv.py $
# Last committed:     $Revision: 587900 $; $Author: vvuik $; $Date: 2020-03-27 09:20:20 +0100 (vr, 27 mrt 2020) $
#
# -*- coding: utf-8 -*-

"""
Created on Thu Aug  1 14:23:20 2019

@author: vuik

Modifications:
22-08-2019:	fmaxS changed from 1.0 into 2.0 Hz
30-08-2019:	correction for pressure attenuation modified into Neumeier method

calculate directional spectra from p,u,v measurements
using maximum entropy method (MEM)

entropy is a measure for number of ways to realize the considered state of a system. 
the system with maximum entropy is the most probable system.

p: surface elevation signal
u: cross-shore velocity signal
v: alongshore velocity signal

output
Hm0: significant wave height
vy: frequency distribution
theta: directional distribution
S: spectral densities

Based on: Ad Reniers 5-8-1999, MEM_dir.m (Matlab)
Method by Lygre & Krogstad 1986, adapted by Cees de Valk
Converted to python and adapted by Vincent Vuik, 29-5-2019

"""

import numpy as np
from scipy.optimize import minimize
import pandas as pd
from jspect import jspect
from disper import disper
from fungrad import fungrad
from calcmoments import calcmoments

def golf_MEMpuv(_p, _u, _v, sensorlevel, bedlevel):
    
    g = 9.81
    rho = 1000
    
    p = (_p*100/g/rho).tolist()
    u = _u.tolist()
    v = _v.tolist()
    
    zIns = sensorlevel # height of instrument (m+NAP)
    zBot = bedlevel   # bottom height near instrument (m+NAP)
    
    elev  = zIns-zBot # distance from instrument to bottom
    depth = np.mean(p)+elev # average water depth at instrument
    hmean = depth+bedlevel # GEM_WATHTE in Aquo: average water level (m+NAP) = mean(p)+zIns

    dt    = 1/4  # sampling interval (at 4 Hz: 1/4 sec)
    fmin  = 0.10 # minimum frequency for calculating spectral parameters (100-1500 mHz)
    fmax  = 1.50 # maximum frequency for calculating spectral parameters (100-1500 mHz)
    
    # determine cross-spectral densities (co-spectra only)
    # cross spectra, for p,u: 1st column is p,p, 2nd u,u, 3rd p,u
    # nsample = 2**12 # default
    nsample = 2**7 # Robert: opletten! 2^7
    F,P = jspect(p,p,nsample,dt,DW='hann',OVERLAP=1,DETREND=1)
    Ppp = np.real(P[2]) # PHI(:,1): p,p
    df  = F[1]-F[0]
    F,P = jspect(p,u,nsample,dt,DW='hann',OVERLAP=1,DETREND=1)
    Ppu = np.real(P[2]) # PHI(:,2): p,u
    F,P = jspect(p,v,nsample,dt,DW='hann',OVERLAP=1,DETREND=1)
    Ppv = np.real(P[2]) # PHI(:,3): p,v
    F,P = jspect(u,u,nsample,dt,DW='hann',OVERLAP=1,DETREND=1)
    Puu = np.real(P[2]) # PHI(:,5): u,u
    F,P = jspect(v,v,nsample,dt,DW='hann',OVERLAP=1,DETREND=1)
    Pvv = np.real(P[2]) # PHI(:,4): v,v
    F,P = jspect(u,v,nsample,dt,DW='hann',OVERLAP=1,DETREND=1)
    Puv = np.real(P[2]) # PHI(:,6): u,v
    
    # mean direction: cross spectra p,v divided by p,u
    theta_m = np.arctan(Ppv/Ppu)
    
    # construct normalised spectral densities
    P_N1 = np.ones(np.size(F))
    P_N2 = Ppu/np.sqrt(Ppp)/np.sqrt(Puu+Pvv)
    P_N3 = Ppv/np.sqrt(Ppp)/np.sqrt(Puu+Pvv)
    P_N4 = (Puu-Pvv)/(Puu+Pvv) 
    P_N5 = 2*Puv/(Puu+Pvv)
    
    Ppp_MAX = np.max(Ppp)
    
    # compute spectrum
    # construct theta-array
    ntheta  = 64
    thetmax = np.pi
    dtheta  = 2*thetmax/ntheta
    theta   = np.arange(start=-thetmax,stop=thetmax,step=dtheta)
    
    # construct fourier components
    qk1_1 = np.ones(np.size(theta))
    qk1_2 = np.cos(theta)
    qk1_3 = np.sin(theta)
    qk1_4 = np.cos(2*theta)
    qk1_5 = np.sin(2*theta)
    qk1   = np.column_stack((qk1_1,qk1_2,qk1_3,qk1_4,qk1_5))
    
    # search options
    maxiter = 100
    
    # obtain f-ky spectral estimate for nl frequencies
    df    = F[1]-F[0]
    fmaxS = 2.0 # Vincent: maximum frequency for 2D spectrum: changed from 1.0 to 2.0 Hz for short waves in Markermeer/IJsselmeer
    nl    = int(np.floor(fmaxS/df))
    
    # first row with zeros for zero frequency
    S = np.zeros(ntheta)
    
    ic   = 1
    for l in np.arange(2,nl+1): # per non-zero frequency
        P_N  = [P_N1[l-1],P_N2[l-1],P_N3[l-1],P_N4[l-1],P_N5[l-1]]
        P_Ni = np.conj(P_N) # complex conjugate of normalized spectral densities
        ic   = ic + 1 # same as l?
        
        if Ppp[l-1]/Ppp_MAX > 0.0001: # only calculate if significant energy is present
            
            mu0 = np.zeros(5) # starting point
            
            # find minumum of cost function for 2D MEM estimator, as a function of mu
            res = minimize(fungrad, jac=True, method='BFGS', x0=mu0, args=(P_Ni,ntheta,dtheta,qk1), options={'maxiter':maxiter})
            mu1 = res.x
            
            # construct directional spreading function
            D = np.exp(-(np.inner(qk1,mu1))) # original code: matrix with size of S
        else:
            D = np.zeros(ntheta) 
        
        # apply pressure attenuation correction, similar to Neumeier
        f        = F[l-1]
        k        = disper(2*np.pi*f,depth)
        KptFull  = np.cosh(k*depth)/(np.cosh(k*elev))  
        fcorrmin = 0.05
        fcorrmax = 1.00 # default in Neumeier toolbox: 0.33 Hz. Changed into 1.00 Hz for low frequency waves in lakes (VV)
        maxcorr  = 5 # 5 in terms of pressure is 5^2=25 in terms of variance density (trans = Kpt**2)
        fvals    = [0,0.75*fcorrmin,fcorrmin,fcorrmax,1.25*fcorrmax,10] # linear tapering of maximum value (VV)
        maxvals  = [1,1,maxcorr,maxcorr,1,1]
        KptMax   = np.interp(f,fvals,maxvals)
        Kpt      = np.min([KptFull,KptMax])
        trans    = Kpt**2
        
        # original code in MEMpuv:
        # trans = (np.cosh(k*depth)/(np.cosh(k*elev)))**2
        # trans = np.min([trans,9]) # 9 = maxcorr**2
    
        # determine 2D variance density spectrum
        S     = np.vstack((S,Ppp[ic-1]*D*trans)) 
        

        
        
        
    # analyse 1D spectrum
    fx    = F[:nl]
    vy    = np.trapz(S,theta)
    ef    = np.where(np.logical_and(fx>=fmin,fx<=fmax))
    fsel  = fx[ef]
    vsel  = vy[ef]
    em    = np.argmax(vsel) 
    fp    = fsel[em] # peak frequency								
    Tp    = 1/fp # peak period	
    mom   = calcmoments(fx,vy,fmin=fmin,fmax=fmax)
    m0    = float(mom['m0'])
    Hm0   = float(4*np.sqrt(mom['m0'])) # Hm0 = 4*m0^0.5
    Tm01  = float(mom['m0']/mom['m1']) # Tm01 = m0/m1
    Tm02  = float(np.sqrt(mom['m0']/mom['m2'])) # Tm02 = (m0/m2)^0.5
    Tmm10 = float(mom['mm1']/mom['m0']) # Tm-1,0 = (m-1/m0)
    
    # determine mean wave direction and directional spreading, Kuik et al. 1988
    # see SWAN user manual for definitions
    # theta0 and dspr only for range [fmin,fmax]
    sT     = np.sin(theta)
    cT     = np.cos(theta)
    idsel  = np.where(np.logical_and(fx>=fmin, fx<=fmax))
    m0Sin  = np.trapz((np.dot(S[idsel,:], sT))*dtheta,fx[idsel])
    m0Cos  = np.trapz((np.dot(S[idsel,:], cT))*dtheta,fx[idsel])
    theta0 = np.rad2deg(np.arctan2(m0Sin, m0Cos)) % 360 # cartesian, degrees
    dspr2  = 2*(1-np.sqrt((m0Sin/m0)**2+(m0Cos/m0)**2))
    dspr   = np.rad2deg(np.sqrt(dspr2)) # directional spreading in degrees 
    
    li = [hmean, Hm0, Tp, Tm01, Tm02, Tmm10, theta0[0], dspr[0]]

    return li, vy, fx, S



